/* final_hps.c
 * ECE 5760
 * Joshua Diaz, Ryan Hornung, Adam Weld
 * jd794, rmh286, aw698 @cornell.edu
 * Cornell University
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdint.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/mman.h>
#include <sys/time.h>
#include <math.h>
#include "helpers.h"

// Main bus PIO
#define FPGA_AXI_BASE 0xC0000000
#define FPGA_AXI_SPAN 0x00001000

// PIO base values
#define PIO_PROGRAM_BASE 0x00000000 
#define PIO_CONTROL_BASE 0x00000010 
#define PIO_FREQUENCY_BASE 0x00000020
#define PIO_FPGA_OUTPUT_BASE 0x00000030

//Parameters for algorithm
#define POPULATION_SIZE 128
#define NUM_OUTPUT_SAMPLES 200

// main axi bus base
void *h2p_virtual_base;

// Pointers to FPGA PIO
volatile unsigned int *h2p_axi_prog_ptr = NULL;
volatile unsigned int *h2p_axi_control_ptr = NULL;
volatile unsigned int *h2p_axi_frequency_ptr = NULL;
volatile unsigned int *h2p_axi_fpgaout_ptr = NULL;

// /dev/mem file id
int fd;

// Population of genomes
struct genome population[POPULATION_SIZE];

// coming out of fpga
// [low_signal, high_signal]
// 200 samples of fpga output for each of low and high signal
// for each individual in the population
int fpga_out[POPULATION_SIZE][2][NUM_OUTPUT_SAMPLES];

// File to save most fit individual into
FILE *fit_file;

int main(void){
    // Iteration vars
    int i, j;
    int gen_count = 0;

    // 100kHz timer
    const struct timespec timer = {.tv_sec = 0, .tv_nsec = 10000};

    // Get FPGA addresses
    if((fd = open("/dev/mem", (O_RDWR | O_SYNC))) == -1){
        printf("ERROR: could not open \"/dev/mem\"...\n");
        return 1;
    }

    // Get virtual addr that maps to the physical
    h2p_virtual_base = mmap(NULL, FPGA_AXI_SPAN, (PROT_READ | PROT_WRITE),
                                MAP_SHARED, fd, FPGA_AXI_BASE); 

    if(h2p_virtual_base == MAP_FAILED){
        printf("ERROR: mmap() failed");
        close(fd);
        return 1;
    }

    // Get addresses mapping to PIO
    h2p_axi_prog_ptr = (unsigned int *)(h2p_virtual_base + PIO_PROGRAM_BASE);
    h2p_axi_control_ptr = (unsigned int *)(h2p_virtual_base + PIO_CONTROL_BASE);
    h2p_axi_frequency_ptr = (unsigned int *)(h2p_virtual_base + PIO_FREQUENCY_BASE);
    h2p_axi_fpgaout_ptr = (unsigned int *)(h2p_virtual_base + PIO_FPGA_OUTPUT_BASE);

    /** INITIALIZE RANDOM POPULATION **/
    for(i = 0; i < POPULATION_SIZE; i++){
        // Set random seed
        time_t seconds;
        time(&seconds);
        srand((unsigned int)seconds);
        // Populate
        for(j = 0; j < WORDS_PER_GENOME; j++){
            // RAND_MAX == 2^31
            // left shift 1 and make LSB random for random full 32 bits
            uint32_t LSB = (rand() % 2);
            population[i].data[j] = (uint32_t)((rand() << 1) | LSB);
        }      
    }

    while(1){
        /**
        FOR EACH SPECIES
            (1) Load prog to FPGA
            (2) set sig gen to lower freq
            (3) read output of FPGA
            (4) set sig to higher freq
            (5) read output of fpga
            (6) determine fitness
        **/
        /** EVALUATE FITNESS OF EACH INDIVIDUAL, ON THE FGPA **/
        for(i = 0; i < POPULATION_SIZE; i++){
            // Load genome to FPGA
            for(j = 0; j < WORDS_PER_GENOME; j++){
                *h2p_axi_control_ptr = 1;
                *h2p_axi_prog_ptr = population[i].data[j];
                *h2p_axi_control_ptr = 0;
            }
            // Set signal generator to lower frequency
            *h2p_axi_frequency_ptr = 0;
            // wait just a bit for output to normalize
            for(j = 0; j < 100; j++){
                nanosleep(&timer);
            }
            // Sample output of FPGA @ 100kHz, 200 samples
            for(j = 0; j < NUM_OUTPUT_SAMPLES; j++){
                fpga_out[i][0][j] = *h2p_axi_fpgaout_ptr;
                nanosleep(&timer, NULL);
            }
            // wait just a bit for output to normalize
            for(j = 0; j < 100; j++){
                nanosleep(&timer);
            }
            // Set signal generator to higher frequency
            *h2p_axi_frequency_ptr = 1;
            // Sample output of FPGA @ 100kHz, 200 samples
            for(j = 0; j < NUM_OUTPUT_SAMPLES; j++){
                fpga_out[i][1][j] = *h2p_axi_fpgaout_ptr;
                nanosleep(&timer, NULL);
            }

            // Determine fitness
            // Fitness here is the product of (1) the percentage of 
            // low-signal outputs that match desired out and (2) the
            // percentage of high-signal outputs that match the desired out
            float temp1 = 0.0, temp2 = 0.0;
            for(j = 0; j < NUM_OUTPUT_SAMPLES; j++){
                temp1 += (float)(1 - fpga_out[i][0][j]);
                temp2 += (float)(fpga_out[i][1][j]);
                population[i].fitness = 
                    (temp1/(float)NUM_OUTPUT_SAMPLES) * (temp2/(float)NUM_OUTPUT_SAMPLES);
            }
        }

        /**
        CREATE NEXT GEN
            (7) Let top few survive to next gen
            (8) Mutations on others
            (9) Crossing over on others
        **/
        // Sort population by fitness
        qsort(population, POPULATION_SIZE, sizeof(struct genome), cmp_genome);
        // Perform random mutations with 1 percent chance
        // on all but top 20 genomes
        // Mutations are simply the toggling of a single bit
        for(j = 0; j < (POPULATION_SIZE - 20); j++){
            // Set random seed
            time_t seconds;
            time(&seconds);
            srand((unsigned int)seconds);
            if((rand() % 101) <= 1){
                int word = (rand() % (WORDS_PER_GENOME));
                int bit = (rand() % 32);
                population[j].data[word] ^= 1<<bit;
            }
        }
        // Each genome (except top 20) has a 1 percent chance 
        // at performing a crossover with another random genome
        // Crossovers are the swapping of a random word from the
        // first genome and a random word of the second
        for(j = 0; j < (POPULATION_SIZE - 20); j++){
            // Set random seed
            time_t seconds;
            time(&seconds);
            srand((unsigned int)seconds);
            if((rand() % 101) <= 1){
                int other_guy = (rand() % 80);
                int word_1 = (rand() % (WORDS_PER_GENOME));
                int word_2 = (rand() % (WORDS_PER_GENOME));
                swap(&(population[j].data[word_1]), &(population[other_guy].data[word_2]));
            }
        }

        /**
        Every 1000 gens, save most fit individual to a txt
        **/
        if((gen_count % 1000) == 0){
            fit_file = fopen("most_fit.txt", "w");
            if(fit_file == NULL){
                printf("Unable to write to \"most_fit.txt\"");
                return 1;
            } 
            fwrite(population[POPULATION_SIZE-1].data, sizeof(uint32_t), WORDS_PER_GENOME, fit_file);
            fclose(fit_file);
        }
        
        gen_count++;
        printf("starting generation %d\n", gen_count);
    }
    return 0;
}

